import 'dart:convert';
import 'dart:math';
import 'dart:typed_data';

import 'package:encrypt/encrypt.dart';
import 'package:crypto/crypto.dart';

/// Class to make the most common cases in cryptography straight forward
///
/// Uses the package encrypt for encryption and crypto for hashing
///
/// Encryption also generates a (salted) hash of the data to validate decryption correctness
///
/// I'm not deep into cryptography and this code may NOT be cryptographically safe.
/// Only use it on your own risk. (But if you see weaknesses in this approach, please email me)
class Crypter {
  /// Encrypt a custom Uint8List (not null) with a key (not null)
  ///
  /// The result consists of 256 bit salt, 256 bit hash, and the encrypted data
  /// The input values for aes256 (mode sic) are the first 128 bits of the salt
  /// the hash is generated with sha256 from the data (salt gets appended to data)
  static Uint8List encrypt(Uint8List data, String key) {
    assert(data != null);
    assert(key != null);
    var salt = _random(32);
    var keyHash = _keyHash(key);
    var iv = IV(salt.sublist(0, 16));
    var hash = sha256.convert(_merge([data, salt])).bytes;
    data = AES(Key(keyHash)).encrypt(data, iv: iv).bytes;
    return _merge([salt, hash, data]);
  }

  /// Decrypt a previously encrypted Uint8List (not null) with a key (not null)
  ///
  /// Returns null if padding is wrong or if the validation with the hash fails
  static Uint8List decrypt(Uint8List data, String key) {
    assert(data != null);
    assert(key != null);
    var salt = data.sublist(0, 32);
    var keyHash = _keyHash(key);
    var iv = IV(salt.sublist(0, 16));
    var hash = data.sublist(32, 64);
    data = data.sublist(64);
    try {
      data = AES(Key(keyHash)).decrypt(Encrypted(data), iv: iv);
    } catch (_) {
      return null;
    }
    var hash2 = sha256.convert(_merge([data, salt])).bytes;
    bool success = _equals(hash, hash2);
    return success ? data : null;
  }

  /// Generates salt (Random.secure) and hash for some data
  ///
  /// returns the 256 bit sha256 hash appended to the 256 bit salt (generated with Random.secure())
  static Uint8List hash(Uint8List data) {
    assert(data != null);
    var salt = _random(32);
    data = _merge([data, salt]);
    return _merge([salt, sha256.convert(data).bytes]);
  }

  /// Checks wether some data has the same hash as some previously hashed data
  ///
  /// The hash must consist of 256 bit salt and 256 bit sha256 hash of the previous data
  static bool checkHash(Uint8List data, Uint8List hash) {
    assert(data != null);
    assert(hash != null);
    var salt = hash.sublist(0, 32);
    hash = hash.sublist(32, 64);
    data = _merge([data, salt]);
    return _equals(sha256.convert(data).bytes, hash);
  }

  static bool _equals(Uint8List list1, Uint8List list2) {
    if (list1 == null) {
      return list2 == null;
    }
    if (list1.length != list2.length) {
      return false;
    }
    for (int i = 0; i < list1.length; i++) {
      if (list1[i] != list2[i]) {
        return false;
      }
    }
    return true;
  }

  static Uint8List _keyHash(String key) {
    assert(key != null);
    return sha256.convert(utf8.encode(key)).bytes;
  }

  static Uint8List _merge(List<Uint8List> lists) {
    assert(lists != null);
    int length = 0;
    for (var list in lists) {
      length += list.length;
    }
    var result = Uint8List(length);
    int index = 0;
    for (var list in lists) {
      var end = index + list.length;
      result.setRange(index, end, list);
      index = end;
    }
    return result;
  }

  static Uint8List _random(int length) {
    assert(length != null);
    assert(length >= 0);
    var random = Random.secure();
    var result = Uint8List(length);
    for (int i = 0; i < result.length; i++) {
      result[i] = random.nextInt(256);
    }
    return result;
  }
}
